import boto3
import botocore
import os
import shutil
import zipfile
from boto3.s3.transfer import S3Transfer
from vulcan.aws.services._session import AWSSession
from vulcan.aws._common import ls, which
from vulcan.aws._common import execute, execute_in_docker
from subprocess import check_call, CalledProcessError
from datetime import datetime

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


class AWSLambda(AWSSession):

    def __init__(self, **kwargs):
        super(
            self.__class__,
            self
        ).__init__(kwargs['profile_name'], kwargs.get('region_name', None))

        self.profile_name = kwargs['profile_name']
        self.function_name = kwargs['function_name']

        if 'pip_requirements' in kwargs:
            self.pip_requirements = kwargs['pip_requirements']
        else:
            self.pip_requirements = list()

        self.s3_filename = kwargs['s3_filename']

        url = urlparse(kwargs['s3_uri'])
        self.s3_bucket = url.netloc
        self.s3_key = url.path

        if self.s3_key.endswith('/'):
            self.s3_key = "%s%s" % (
                self.s3_key, os.path.basename(self.s3_filename))

        if self.s3_key.startswith('/'):
            self.s3_key = self.s3_key[1:]

    def validate(self):
        self.install_deps(['--user'])
        errors = list()
        files = ls('./src/main/python', '*.py')

        if not which('pylint'):
            print('pylint command not found,'
                  ' skipping code validation via PyLint')
            return True

        if len(files) == 0:
            print("No python files for validation")
            return True
        else:
            print("Validating files via pylint: %s" % files)

        for filename in files:
            try:
                execute('pylint', '--extension-pkg-whitelist=pymssql',
                        '--errors-only', filename)
            except Exception as e:
                errors.append(filename)

        if len(errors) > 0:
            print("There are errors in python files: %s" % errors)
            raise Exception(
                "Lambda validation error. Check output for details.")

        return True

    def package(self, **kwargs):
        print('[ Packaging lambda deployment package ]')
        shutil.rmtree('target/distrib/', ignore_errors=True)

        files = ls('src/main/python/', '*.py')
        cwd = os.getcwd()
        print('[ Files ]')
        for name in files:
            print(name[len(cwd) + 1:])

        shutil.copytree('src/main/python/', 'target/distrib/',
                        ignore=shutil.ignore_patterns('*.pyc', 'tmp*'))

        self.install_deps(
            ['--target', 'target/distrib/'],
            docker_image=kwargs.get('docker_image', None)
        )

        zipf = zipfile.ZipFile(
            'target/{}'.format(self.s3_filename), 'w', zipfile.ZIP_DEFLATED)
        # with safe_cd('target/distrib/')
        basedir_len = len(os.path.abspath(
            os.path.join(os.getcwd(), 'target/distrib')))
        for name in ls('target/distrib/'):
            zipf.write(name, name[basedir_len:])
        zipf.close()

        return True

    def install_deps(self, pip_args=None, **kwargs):
        print('[ Installing lambda dependencies ]')
        print('[ Dependencies ]')

        args = ['install', '--upgrade']

        if pip_args:
            if type(pip_args) != list:
                raise "pip_args argument should be of a list() type"

            for pip_arg in pip_args:
                if pip_arg == '--user' and os.getenv('VIRTUAL_ENV', False):
                    # skip --user when inside virtualenv
                    pass
                else:
                    args.append(pip_arg)

        for req in self.pip_requirements:
            if kwargs['docker_image']:
                print('Installing {req} via docker image {docker}'.format(
                    req=req,
                    docker=kwargs['docker_image'])
                )
                execute_in_docker(kwargs['docker_image'], 'pip', args, req)
            else:
                print('Installing {}'.format(req))
                execute('pip', args, req)

        if not self.pip_requirements or len(self.pip_requirements) == 0:
            print("Your lambda doesn't have any pip dependencies")

    def upload(self):
        s3 = self.client('s3')
        lambda_package = os.path.normpath(os.path.join(
            os.getcwd(), 'target/', self.s3_filename))
        print("Uploading %s to temporary location s3://%s/%s" %
              (lambda_package, self.s3_bucket, self.s3_key))
        S3Transfer(s3).upload_file(
            lambda_package,
            self.s3_bucket,
            self.s3_key,
            extra_args={'ACL': 'bucket-owner-full-control'}
        )

    def update_code(self, **kwargs):
        lambdas = self.client('lambda')
        if kwargs and 'function_name' in kwargs:
            function_name = kwargs['function_name']
        else:
            function_name = self.function_name
        print("Updating function {name} "
              "code from s3://{bucket}/{key}".format(
                  name=function_name,
                  bucket=self.s3_bucket,
                  key=self.s3_key))
        resp = lambdas.update_function_code(
            FunctionName=function_name,
            S3Bucket=self.s3_bucket,
            S3Key=self.s3_key,
            Publish=False
        )

    def update_dev_alias(self):
        lambdas = self.client('lambda')
        print('Updating function {} DEV alias to version $LATEST'.format(
            self.function_name))
        lambdas.update_alias(
            FunctionName=self.function_name,
            FunctionVersion='$LATEST',
            Name='DEV',
            Description='Updated by pynt at {} UTC'.format(
                datetime.utcnow().strftime("%Y-%b-%d %H:%M:%S"))
        )

        pass

    def update_prod_alias(self):
        lambdas = self.client('lambda')
        resp = lambdas.publish_version(
            FunctionName=self.function_name,
        )

        print('Updating function {} PROD alias to version {}'.format(
            self.function_name, resp['Version']))
        lambdas.update_alias(
            FunctionName=self.function_name,
            FunctionVersion=resp['Version'],
            Name='PROD',
            Description='Updated by pynt at {} UTC'.format(
                datetime.utcnow().strftime("%Y-%b-%d %H:%M:%S"))
        )

    def test_local(self, event={}, **kwargs):
        lambdas = self.client('lambda')
        if kwargs and 'function_name' in kwargs:
            function_name = kwargs['function_name']
        else:
            function_name = self.function_name
        resp = lambdas.get_function(
            FunctionName=function_name
        )

        sep = resp['Configuration']['Handler'].rfind('.')
        py_file_name = resp['Configuration']['Handler'][:sep]
        py_method_name = resp['Configuration']['Handler'][sep + 1:]

        import logging
        logging.basicConfig()
        logger = logging.getLogger('lambda')
        logger.setLevel('DEBUG')

        import sys
        import platform
        if platform.system() == 'Windows':
            sys.path.append('src\\main\\python')
        else:
            sys.path.append('src/main/python')

        import importlib
        mod = importlib.import_module(py_file_name)
        result = getattr(mod, py_method_name)(event, {})



from subprocess import check_call, CalledProcessError
import sys
import os
import fnmatch
import contextlib
import random
import json
from datetime import date, datetime


@contextlib.contextmanager
def safe_cd(path):
    starting_directory = os.getcwd()
    try:
        os.chdir(path)
        yield
    finally:
        os.chdir(starting_directory)


def execute(script, *args):
    popen_args = [script]

    for arg in args:
        if type(arg) == list:
            popen_args.extend(arg)
        else:
            popen_args.append(arg)

    try:
        return check_call(popen_args, shell=False)
    except CalledProcessError as ex:
        # print(ex)
        # sys.exit(ex.returncode)
        raise ex
    except Exception as ex:
        print('Error running script '
              '{} with args {}: {}'.format(script, args, ex))
        sys.exit(1)


def execute_in_docker(script, *args):
    popen_args = [
        'docker', 'run', '-v',
        "{pwd}:/app".format(pwd=os.getcwd()),
        '-ti', '--rm', '-w', '/app'
    ]
    popen_args.append(script)

    for arg in args:
        if type(arg) == list:
            popen_args.extend(arg)
        else:
            popen_args.append(arg)

    try:
        return check_call(popen_args, shell=False)
    except CalledProcessError as ex:
        # print(ex)
        # sys.exit(ex.returncode)
        raise ex
    except Exception as ex:
        print('Error running script '
              '{} with args {}: {}'.format(script, args, ex))
        sys.exit(1)


def ls(walk_dir, pattern='*'):
    walk_dir = os.path.abspath(os.path.join(os.getcwd(), walk_dir))
    if not os.path.isdir(walk_dir):
        raise Exception("Wrong dir path: %s" % walk_dir)

    result = list()
    for dirpath, dirnames, files in os.walk(walk_dir):
        for filename in [f for f in files if fnmatch.fnmatch(f, pattern)]:
            full_filename = os.path.join(dirpath, filename)
            result.append(full_filename)

    return result


def which(program):
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None


def generatePassword(**kwargs):
    _kwargs = {
        'len': 8,
        'symbols': "!@#$%^&*()_+-=[]{}",
        'digits': '0123456789',
        'upper_count': 2,
        'digits_count': 1,
        'symbols_count': 1
    }
    _kwargs.update(kwargs)

    _kwargs['len'] = max(_kwargs['len'],
                         1 +
                         int(_kwargs.get('upper_count')) +
                         min(len(_kwargs.get('digits')),
                             _kwargs.get('digits_count')) +
                         min(len(_kwargs.get('symbols')),
                             _kwargs.get('symbols_count'))
                         )

    rnd = random.sample(
        range(0, int(_kwargs.get('len'))),
        int(_kwargs.get('upper_count')) +
        min(len(_kwargs.get('digits')), _kwargs.get('digits_count')) +
        min(len(_kwargs.get('symbols')), _kwargs.get('symbols_count'))
    )

    lowcase_letters = "abcdefghijklmnopqrstuvwxyz"
    upcase_letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

    # generate base password string using lowcase alphabet
    result = "".join([random.SystemRandom().choice(lowcase_letters)
                      for _ in xrange(int(_kwargs.get('len')))])

    # put uppercase if any
    if _kwargs.get('upper_count') and _kwargs.get('upper_count') > 0:
        for x in range(0, _kwargs.get('upper_count')):
            rndUpcaseLetter = random.randint(0, len(upcase_letters) - 1)
            result = replaceIdx(result, rnd.pop(), upcase_letters[
                rndUpcaseLetter])

    if _kwargs.get('digits_count') and _kwargs.get('digits_count') > 0:
        for x in range(0, _kwargs.get('digits_count')):
            rndDigitIdx = random.randint(0, len(_kwargs.get('digits')) - 1)
            result = replaceIdx(result, rnd.pop(), _kwargs.get('digits')[
                rndDigitIdx])

    if _kwargs.get('symbols_count') and _kwargs.get('symbols_count') > 0:
        for x in range(0, _kwargs.get('symbols_count')):
            rndSymbolIdx = random.randint(0, len(_kwargs.get('symbols')) - 1)
            result = replaceIdx(result, rnd.pop(), _kwargs.get('symbols')[
                rndSymbolIdx])

    return result


def replaceIdx(s, idx, char):
    return s[:idx] + char + s[idx + 1:]


_pyntexit_registered = False


def _pyntexit_register():
    global _pyntexit_registered
    if not _pyntexit_registered:
        import atexit
        atexit.register(_pyntexit)
        _pyntexit_registered = True


def _pyntexit():
    import datetime
    print('Script exited at {}'.format(
        datetime.datetime.now().strftime("%m/%d/%Y %H:%M:%S")))
    import platform
    if platform.system() == 'Darwin':
        execute('osascript', '-e',
                'display notification "Script finished" with title "Pynt"')


def json_serial(obj):
    """JSON serializer for objects not serializable by default json code"""

    if isinstance(obj, (datetime, date)):
        return obj.isoformat()
    raise TypeError("Type %s not serializable" % type(obj))


def dump(obj):
    print(json.dumps(obj, indent=1, default=json_serial))



import boto3
import threading


class AWSSession(object):

    def __init__(self, profile_name, region_name=None):
        self.profile_name = profile_name
        self.region_name = region_name
        self._thread_local = threading.local()
        self._thread_local.boto3_sessions = dict()
        self._thread_local.cache = dict()

    def __session(self):
        sessions = self._thread_local.boto3_sessions
        session_key = '{}.{}'.format(
            self.region_name,
            self.profile_name
        )

        if session_key not in sessions:
            # print("[{}] session.Session(): key={}".format(threading.get_ident(), session_key))
            if self.profile_name in boto3.session.Session().available_profiles:
                if self.region_name:
                    sessions[session_key] = boto3.session.Session(
                        profile_name=self.profile_name,
                        region_name=self.region_name
                    )
                else:
                    sessions[session_key] = boto3.session.Session(
                        profile_name=self.profile_name)
            else:
                sessions[session_key] = boto3.session.Session()

        return sessions[session_key]

    def client(self, name):
        return self.__session().client(name)

    def resource(self, name):
        return self.__session().resource(name)

    def cache(self, key, value=None):
        if value:
            cache = self._thread_local.cache
            cache[key] = value
            return value
        else:
            cache = self._thread_local.cache
            return cache.get(key, None)



"""
Amazon AWS boto3 helper libs
"""

from vulcan.aws.services._iam import AWSIAM
from vulcan.aws.services._session import AWSSession
from vulcan.aws.services._cloudformation import AWSCloudFormation
from vulcan.aws.services._lambda import AWSLambda
from vulcan.aws.services._s3 import AWSS3
from vulcan.aws.services._cloudfront import AWSCloudFront
from vulcan.aws.services._ec2 import AWSEC2
from vulcan.aws.services._logs import AWSLogs
from vulcan.aws.services._acm import AWSACM
from vulcan.aws._common import generatePassword
from vulcan.aws._common import safe_cd

import pkgutil
__path__ = pkgutil.extend_path(__path__, __name__)

__all__ = [
    'AWSSession', 'AWSLambda', 'AWSCloudFormation', 'AWSIAM',
    'AWSS3', 'AWSCloudFront', 'AWSEC2', 'AWSLogs', 'AWSACM',
    'generatePassword', 'safe_cd'
]



import boto3
from vulcan.aws.services._session import AWSSession


class AWSLogs(AWSSession):

    def __init__(self, **kwargs):
        super(
            self.__class__,
            self
        ).__init__(kwargs['profile_name'], kwargs.get('region_name', None))

    def group_exists(self, **kwargs):
        if 'group_name' not in kwargs:
            raise Exception('Argument missing: group_name')

        resp = self.client('logs').describe_log_groups(
            logGroupNamePrefix=kwargs.get('group_name')
        )

        result = False
        for group in resp['logGroups']:
            if group['logGroupName'] == kwargs['group_name']:
                result = True

        return result



import json
from vulcan.aws.services._session import AWSSession
from vulcan.aws._common import json_serial


class AWSEC2(AWSSession):

    def __init__(self, **kwargs):
        super(
            self.__class__,
            self
        ).__init__(kwargs['profile_name'], kwargs.get('region_name', None))

    def generate_ssh_config(self, **kwargs):
        filters = list()
        filters.append({
            'Name': 'instance-state-name',
            'Values': ['running']
        })
        if 'tags' in kwargs:
            for tag in kwargs.get('tags'):
                if 'name' in tag and 'value' in tag:
                    filters.append({
                        'Name': 'tag:{}'.format(tag.get('name')),
                        'Values': [tag.get('value')]
                    })
                elif 'name' in tag and 'value' not in tag:
                    filters.append({
                        'Name': 'tag-key',
                        'Values': [tag.get('name')]
                    })
                elif 'name' not in tag and 'value' in tag:
                    filters.append({
                        'Name': 'tag-value',
                        'Values': [tag.get('value')]
                    })
                else:
                    raise Exception('Wrong value for tags element:'
                                    ' {}'.format(json.dumps(tag)))

        resp = self.client('ec2').describe_instances(
            Filters=filters
        )

        print(json.dumps(filters, indent=2, default=json_serial))
        print(json.dumps(resp, indent=2, default=json_serial))



import boto3
import uuid
from vulcan.aws.services._session import AWSSession


class AWSACM(AWSSession):

    def __init__(self, **kwargs):
        super(
            self.__class__,
            self
        ).__init__(kwargs['profile_name'], kwargs.get('region_name', None))

    def find_cert_arn(self, **kwargs):
        cert_arn = None

        if 'domain_name' not in kwargs:
            raise Exception('Argument missing: domain_name')

        client = self.client('acm')

        paginator = client.get_paginator('list_certificates')

        page_iterator = paginator.paginate(CertificateStatuses=['ISSUED'])
        for page in page_iterator:
            if 'CertificateSummaryList' in page:
                for cert in page['CertificateSummaryList']:
                    if cert['DomainName'] == kwargs['domain_name']:
                        if cert_arn is not None:
                            raise Exception('Multiple certificates ''with same domain name.')
                        else:
                            cert_arn = cert['CertificateArn']

        return cert_arn

    def request_via_dns(self, **kwargs):
        if 'domain_name' not in kwargs:
            raise Exception('Argument missing: domain_name')

        if 'alternative_names' not in kwargs:
            raise Exception('Argument missing: alternative_names')

        client = self.client('acm')

        resp = client.request_certificate(
            DomainName=kwargs.get('domain_name'),
            SubjectAlternativeNames=kwargs.get('alternative_names'),
            ValidationMethod='DNS',
            IdempotencyToken=uuid.uuid4()
        )

        return resp

    def get_dns_validation_options(self, **kwargs):
        if 'certificate_arn' not in kwargs:
            raise Exception('Argument missing: certificate_arn')

        client = self.client('acm')

        resp = client.describe_certificate(
            CertificateArn=certificate_arn
        )

        return resp['Certificate']['DomainValidationOptions'][0]['ResourceRecord']



import boto3
import botocore
import os
import uuid
import fnmatch
import re
import sh
import mimetypes
import copy
import traceback
import sys
from boto3.s3.transfer import S3Transfer
from vulcan.aws.services._session import AWSSession
from concurrent import futures

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


class AWSS3(AWSSession):

    def __init__(self, **kwargs):
        super(
            self.__class__,
            self
        ).__init__(kwargs['profile_name'], kwargs.get('region_name', None))

    def bucket_exists(self, **kwargs):
        if 'bucket_name' not in kwargs:
            raise Exception('Argument missing: bucket_name')

        s3api = self.client('s3')

        try:
            s3api.head_bucket(Bucket=kwargs.get('bucket_name'))
        except botocore.exceptions.ClientError as err:
            err_msg = err.response['Error']['Message']
            err_code = err.response['Error']['Code']
            if (err_msg != ("An error occurred (403) when calling "
                            "the HeadBucket operation: Forbidden")):
                return False
            else:
                raise Exception("Error calling head_bucket: {}".format(
                    err_msg), sys.exc_info()[2])

        return True

    def copy_file(self, **kwargs):
        if 'bucket_name' not in kwargs:
            raise Exception('Argument missing: bucket_name')

        s3api = self.client('s3')

        file = open(kwargs.get('file'), 'rb')
        print('Uploading {file} to s3://{bucket}{key}'.format(
            file=file,
            bucket=kwargs.get('bucket_name'),
            key=kwargs.get('key'))
        )
        s3api.put_object(
            Bucket=kwargs.get('bucket_name'),
            Key=kwargs.get('key'),
            Body=file
        )
        file.close()

        return

    def sync(self, **kwargs):
        if 'metadata' in kwargs:
            return self._sync_with_metadata(**kwargs)
        else:
            return self._sync_no_metadata(**kwargs)

    def _sync_with_metadata(self, **kwargs):
        s3api = self.client('s3')
        s3res = self.resource('s3')

        result = dict()
        result['updated_files'] = list()

        s3_bucket, s3_key = self._split_s3_url(kwargs['destination'])
        s3_files_list = self._list_s3_files(s3_bucket, s3_key)

        with sh.pushd(kwargs['source']):
            files = list()
            for root, dirnames, filenames in os.walk('.'):
                for filename in filenames:
                    files.append(os.path.join(root, filename))

            files_to_remove = [
                x for x in s3_files_list if './{}'.format(x) not in files]

            result['updated_files'].extend(files_to_remove)

            for key in files_to_remove:
                s3_key_full = '{}{}'.format(s3_key, key)
                futures_results = dict()
                with futures.ThreadPoolExecutor(max_workers=4) as executor:
                    futures_results[executor.submit(self._delete_file, s3api, {
                        's3_bucket': s3_bucket,
                        's3_key': s3_key_full
                    })] = filename

                for future in futures.as_completed(futures_results):
                    if future.exception() is not None:
                        print(
                            'Error/Warning: file={}, error={}'.format(
                                futures_results[future],
                                future.exception()
                            )
                        )

            for filename in files:
                content_type, content_encoding = self._get_content_type(
                    filename)
                metadata_full_obj = copy.deepcopy(
                    self._get_metadata_for_file(filename, kwargs['metadata']))
                http_metadata = dict()

                result['updated_files'].append(filename[1:])

                # print('gmeta:{}'.format(metadata_full_obj))

                if 'Cache-Control' in metadata_full_obj['metadata']:
                    http_metadata['Cache-Control'] =\
                        metadata_full_obj['metadata'].pop(
                        'Cache-Control', None)

                if 'Content-Type' in metadata_full_obj['metadata']:
                    http_metadata['Content-Type'] =\
                        metadata_full_obj['metadata'].pop(
                        'Content-Type', None)
                elif content_type:
                    http_metadata['Content-Type'] = content_type

                if 'Content-Encoding' in metadata_full_obj['metadata']:
                    http_metadata['Content-Encoding'] =\
                        metadata_full_obj['metadata'].pop(
                            'Content-Encoding', None)
                elif content_encoding:
                    http_metadata['Content-Encoding'] = content_encoding

                s3_key_full = '{}{}'.format(s3_key, filename[2:])
                futures_results = dict()
                with futures.ThreadPoolExecutor(max_workers=4) as executor:
                    futures_results[executor.submit(self._upload_file, s3api, {
                        'filename': filename,
                        's3_bucket': s3_bucket,
                        's3_key': s3_key_full,
                        'acl': kwargs['acl'],
                        'pattern': metadata_full_obj['pattern'],
                        'metadata': metadata_full_obj['metadata'],
                        'http-metadata': http_metadata
                    })] = filename

                for future in futures.as_completed(futures_results):
                    if future.exception() is not None:
                        print('Error/Warning: file={}, error={}'.format(
                            futures_results[future],
                            future.exception()
                        )
                        )

        return result

    def _delete_file(self, s3api, params):
        try:
            print('Removing: s3://{}/{}'.format(
                params['s3_bucket'],
                params['s3_key']
            )
            )
            s3api.delete_object(
                Bucket=params['s3_bucket'],
                Key=params['s3_key'],
            )
        except Exception as e:
            print("Warning: file {} delete error: ({}){}".format(
                params['filename'], type(e), str(e)))

    def _upload_file(self, s3api, metadata):
        with open(metadata['filename'], 'rb') as file:
            print(
                'Uploading: {pattern} {filename} ->'
                ' s3://{bucket}/{key}'.format(
                    pattern=metadata['pattern'],
                    filename=metadata['filename'],
                    bucket=metadata['s3_bucket'],
                    key=metadata['s3_key']
                )
            )
            try:
                if 'Cache-Control' in metadata['http-metadata']:
                    s3api.put_object(
                        Body=file,
                        Bucket=metadata['s3_bucket'],
                        Key=metadata['s3_key'],
                        ACL=metadata['acl'],
                        Metadata=metadata['metadata'],
                        ContentType=metadata['http-metadata']
                        .get('Content-Type', 'application/octet-stream'),
                        ContentLength=os.path.getsize(metadata['filename']),
                        CacheControl=metadata['http-metadata']
                        .get('Cache-Control'),
                    )
                else:
                    s3api.put_object(
                        Body=file,
                        Bucket=metadata['s3_bucket'],
                        Key=metadata['s3_key'],
                        ACL=metadata['acl'],
                        Metadata=metadata['metadata'],
                        ContentType=metadata['http-metadata']
                        .get('Content-Type', 'application/octet-stream'),
                        ContentLength=os.path.getsize(metadata['filename'])
                    )
            except Exception as e:
                print("Warning: file {} copy error: ({}){}".format(
                    metadata['filename'], type(e), str(e)))
                traceback.print_exc(file=sys.stderr)

    def _get_metadata_for_file(self, filename, metadatas):
        for metadata in metadatas:
            if metadata.get('pattern') == '.*':
                default_metadata = metadata['metadata']

            if type(metadata.get('pattern')) == str:
                metadata['pattern'] = [metadata['pattern']]

            for pattern in metadata['pattern']:
                if re.search(pattern, filename):
                    return metadata

    def _list_s3_files(self, s3_bucket, s3_key):
        result = list()

        s3api = self.client('s3')
        paginator = s3api.get_paginator('list_objects_v2')
        page_iterator = paginator.paginate(Bucket=s3_bucket, Prefix=s3_key)
        for page in page_iterator:
            if "Contents" in page:
                for key in page["Contents"]:
                    result.append(key['Key'])

        return result

    def _split_s3_url(self, s3_url):
        url = urlparse(s3_url)
        s3_bucket = url.netloc
        s3_key = url.path

        if s3_key.endswith('/'):
            s3_key = "%s%s" % (s3_key, s3_url)

        if s3_key.startswith('/'):
            s3_key = s3_key[1:]

        return (s3_bucket, s3_key)

    def _get_content_type(self, filename):
        return mimetypes.guess_type(filename)



import uuid
from vulcan.aws.services._session import AWSSession


class AWSCloudFront(AWSSession):

    def __init__(self, **kwargs):
        super(
            self.__class__,
            self
        ).__init__(kwargs['profile_name'], kwargs.get('region_name', None))

        self.distribution_id = kwargs.get('distribution_id')

    def invalidate(self, *files):
        cloudfront = self.client('cloudfront')
        resp = cloudfront.create_invalidation(
            DistributionId=self.distribution_id,
            InvalidationBatch={
                'CallerReference': uuid.uuid4().hex,
                'Paths': {
                    'Quantity': len(files),
                    'Items': files
                }
            }
        )



__version__ = "0.2.33"
__license__ = "MIT License"
__website__ = "https://code.exrny.com/opensource/vulcan-aws/"
__download_url__ = ('https://github.com/exrny/vulcan-aws/archive/'
                    '{}.tar.gz'.format(__version__))



import boto3
import botocore
import os
import uuid
import textwrap
from botocore.exceptions import ClientError
from vulcan.aws.services._session import AWSSession
from vulcan.aws._common import generatePassword


class AWSIAM(AWSSession):

    def __init__(self, **kwargs):
        super(
            self.__class__,
            self
        ).__init__(kwargs['profile_name'], kwargs.get('region_name', None))

    def create_access_key(self, **kwargs):
        resp = self.client('iam').create_access_key(
            UserName=kwargs.get('username')
        )

        if 'print_credentials' in kwargs and bool(kwargs['print_credentials']):
            print(textwrap.dedent('''
        [{profile_name}.ci]
        aws_access_key_id = {key_id}
        aws_secret_access_key = {key_secret}
        '''.format(
                  profile_name=self.profile_name,
                  key_id=resp['AccessKey']['AccessKeyId'],
                  key_secret=resp['AccessKey']['SecretAccessKey']
                  )
            ))
        else:
            return resp['AccessKey']

    def create_password(self, **kwargs):
        aws_iam = self.client('iam')

        account_alias = self.get_signin_url()

        login_profile = None

        try:
            aws_iam.delete_login_profile(UserName=kwargs.get('username'))
            login_profile = aws_iam.get_login_profile(
                UserName=kwargs.get('username'))
        except ClientError as e:
            if e.response["Error"]["Code"] == "NoSuchEntity":
                pass
            else:
                raise e

        password = generatePassword()

        if not login_profile:
            aws_iam.create_login_profile(
                UserName=kwargs.get('username'),
                Password=password,
                PasswordResetRequired=True
            )
        else:
            aws_iam.update_profile(
                UserName=kwargs.get('username'),
                Password=password,
                PasswordResetRequired=True
            )

        if 'print_password' in kwargs and bool(kwargs['print_password']):
            print(textwrap.dedent('''
        Signin URL: https://{alias}.signin.aws.amazon.com/console
        Username: {username}
        Password: {password}
        '''.format(
                  alias=account_alias,
                  username=kwargs.get('username'),
                  password=password
                  )
            ))
        else:
            return password

    def get_signin_url(self):
        aws_iam = self.client('iam')
        resp = aws_iam.list_account_aliases()

        if len(resp['AccountAliases']) != 1:
            raise Exception('Account aliases not equal to 1')

        return resp['AccountAliases'][0]



import boto3
import botocore
import os
import uuid
import json
import sys
import re
import time
import textwrap
import threading
from datetime import datetime, timezone
from boto3.s3.transfer import S3Transfer
from vulcan.aws.services._session import AWSSession
from vulcan.aws import shared
from vulcan.aws._common import dump

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse

shared.store['validated_templates'] = list()
LOG_GROUP = 'cfnotify'

STATUS_UPDATE_COMPLETE = ('UPDATE_COMPLETE', 'UPDATE_FAILED', 'UPDATE_ROLLBACK_FAILED',
                          'UPDATE_ROLLBACK_FAILED', 'UPDATE_ROLLBACK_COMPLETE')
STATUS_CREATE_COMPLETE = ('CREATE_COMPLETE', 'CREATE_FAILED')
STATUS_DELETE_COMPLETE = ('DELETE_COMPLETE', 'DELETE_FAILED')
STATUS_ROLLBACK_COMPLETE = ('ROLLBACK_FAILED', 'ROLLBACK_COMPLETE')

STATUS_CF_RUN_COMPLETE = list()
STATUS_CF_RUN_COMPLETE.extend(STATUS_UPDATE_COMPLETE)
STATUS_CF_RUN_COMPLETE.extend(STATUS_CREATE_COMPLETE)
STATUS_CF_RUN_COMPLETE.extend(STATUS_DELETE_COMPLETE)
STATUS_CF_RUN_COMPLETE.extend(STATUS_ROLLBACK_COMPLETE)


class AWSCloudFormation(AWSSession):

    def __init__(self, **kwargs):
        super(self.__class__, self).__init__(
            kwargs.get('profile_name'),
            kwargs.get('region_name', None)
        )
        self.stack_name = kwargs['stack_name']

        if (len(kwargs) == 2 and
                'profile_name' in kwargs and
                'stack_name' in kwargs):
            # Easy service, for lookups only
            self.easy_service = True
        elif (len(kwargs) == 3 and
              'profile_name' in kwargs and
              'stack_name' in kwargs and
              'region_name' in kwargs):
            # Easy service, for lookups only
            self.easy_service = True
        else:
            self.easy_service = False

            self.on_failure = kwargs.get('on_failure', 'ROLLBACK')

            if 'template' in kwargs and type(kwargs['template']) == str:
                self.template = kwargs['template']
            else:
                raise Exception('Missing or wrong parameter: template')

            self.includes = list()
            if 'includes' in kwargs:
                if type(kwargs['includes']) == list:
                    self.includes = kwargs['includes']
                else:
                    raise Exception('Wrong parameter type: '
                                    'includes = {}'.format(
                                        type(kwargs['includes'])))

            self.resources = list()
            if 'resources' in kwargs:
                if type(kwargs['resources']) == list:
                    self.resources = kwargs['resources']
                else:
                    raise Exception('Wrong parameter type: '
                                    'resources = {}'.format(
                                        type(kwargs['resources'])))

            if 'parameters' in kwargs:
                self.parameters = kwargs['parameters']
            else:
                self.parameters = None

            path_to_cloudformation = os.path.abspath(os.path.join(os.getcwd(), './src/cloudformation/'))
            if not os.path.exists(path_to_cloudformation):
                path_to_cloudformation = os.path.abspath(
                    os.path.join(os.getcwd(), './src/main/cloudformation/')
                )

            path_to_resources = os.path.abspath(os.path.join(os.getcwd(), './src/resources/'))
            if not os.path.exists(path_to_cloudformation):
                path_to_resources = os.path.abspath(os.path.join(os.getcwd(), './src/main/resources/'))

            self.template = os.path.abspath(os.path.join(path_to_cloudformation, self.template))
            for idx, template in enumerate(self.includes):
                if not os.path.isabs(template):
                    if template.startswith('./') or template.startswith('../'):
                        self.includes[idx] = os.path.abspath(
                            os.path.join(os.getcwd(), template))
                    else:
                        self.includes[idx] = os.path.abspath(os.path.join(path_to_cloudformation, template))

                if not os.path.isfile(self.includes[idx]):
                    raise Exception("Can't find template file"
                                    " '{}' ({})".format(
                                        template,
                                        self.includes[idx]))

            for idx, file in enumerate(self.resources):
                if not os.path.isabs(file):
                    if file.startswith('./') or file.startswith('../'):
                        self.resources[idx] = os.path.abspath(
                            os.path.join(os.getcwd(), file))
                    else:
                        self.resources[idx] = os.path.abspath(os.path.join(path_to_resources, file))

                if not os.path.isfile(self.resources[idx]):
                    raise Exception("Can't find resource file "
                                    "'{}' ({})".format(
                                        file, self.resources[idx]))

            url = urlparse(kwargs['s3_uri'])
            self.s3_bucket = url.netloc
            self.s3_key = url.path

            if self.s3_key.endswith('/'):
                self.s3_key = "%s%s" % (
                    self.s3_key, os.path.basename(self.template))

            if self.s3_key.startswith('/'):
                self.s3_key = self.s3_key[1:]

    # @property
    # def s3_uri(self):
    #   return self._s3_uri

    def exists(self):
        cloudformation = self.client('cloudformation')
        STACK_EXISTS_STATES = [
            'CREATE_COMPLETE',
            'ROLLBACK_COMPLETE',
            'UPDATE_COMPLETE',
            'UPDATE_ROLLBACK_COMPLETE'
        ]
        try:
            stack = None
            nextToken = None
            while not stack:
                resp = None
                if nextToken:
                    resp = cloudformation.describe_stacks(
                        StackName=self.stack_name, NextToken=nextToken)
                else:
                    resp = cloudformation.describe_stacks(
                        StackName=self.stack_name)

                for stack in resp['Stacks']:
                    if stack['StackStatus'] in STACK_EXISTS_STATES:
                        return True
                if 'NextToken' in stack:
                    nextToken = stack['NextToken']

            return False
        except botocore.exceptions.ClientError as err:
            err_msg = err.response['Error']['Message']
            err_code = err.response['Error']['Code']
            if err_msg != "Stack with id {} does not exist".format(self.stack_name) and \
               err_code != 'ValidationError':
                return False

    def print_outputs(self, **kwargs):
        stack_outputs = self.outputs(**kwargs)

        if len(stack_outputs) == 0:
            print("Stack {} don't have any outputs".format(self.stack_name))
        else:
            print("Stack {} outputs:".format(self.stack_name))
            max_name_len = 0
            for output in stack_outputs:
                max_name_len = max(max_name_len, len(output['OutputKey']))
            for output in stack_outputs:
                print(' '.join([
                    output['OutputKey'].ljust(max_name_len) + ':',
                    output['OutputValue'],
                ]))

    def outputs(self, output_key=None, **kwargs):
        cloudformation = self.client('cloudformation')
        STACK_EXISTS_STATES = [
            'CREATE_COMPLETE',
            'ROLLBACK_COMPLETE',
            'UPDATE_COMPLETE',
            'UPDATE_ROLLBACK_COMPLETE'
        ]

        no_fail = False
        no_cache = False
        print_out = False
        if kwargs:
            no_fail = kwargs.get('no_fail', False)
            no_cache = kwargs.get('no_cache', False)
            print_out = kwargs.get('print', False)

        stack_outputs = None
        nextToken = None
        cache_key = 'cloudformation.outputs.{}.{}.{}'.format(
            self.region_name,
            self.profile_name,
            self.stack_name
        )
        if not no_cache:
            stack_outputs = self.cache(cache_key)

        if not stack_outputs:
            try:
                stack = None
                stack_outputs = list()

                paginator = cloudformation.get_paginator('describe_stacks')
                page_iterator = paginator.paginate(StackName=self.stack_name)
                break_for = False
                for page in page_iterator:
                    for stack in page['Stacks']:
                        if stack['StackStatus'] in STACK_EXISTS_STATES:
                            stack_outputs = stack.get('Outputs', list())
                            self.cache(cache_key, stack_outputs)
                            break_for = True
                            break
                    if break_for:
                        break
            except botocore.exceptions.ClientError as err:
                err_msg = err.response['Error']['Message']
                err_code = err.response['Error']['Code']
                if err_msg != "Stack with id {} does not exist".format(self.stack_name) and \
                   err_code != 'ValidationError':
                    if no_fail:
                        print("Stack with id {} does not exist".format(self.stack_name))
                    else:
                        raise Exception("Stack with id {} does not exist".format(
                            self.stack_name), sys.exc_info()[2])

        if output_key:
            print('AWSCloudFormation.output(): WARING: outputs method is deprecated, use output instead')
            for output in stack_outputs:
                if output['OutputKey'] == output_key:
                    return output['OutputValue']
            print("AWSCloudFormation.output(): " +
                  "Can't find output parameter {} in stack {} under {} profile".format(
                      output_key,
                      self.stack_name,
                      self.profile_name
                  ))

            return None
        else:
            if print_out:
                if len(stack_outputs) == 0:
                    print("Stack {} don't have any outputs".format(self.stack_name))
                else:
                    print("Stack {} outputs:".format(self.stack_name))
                    for output in stack_outputs:
                        print('{key}: {value}'.format(
                            key=output['OutputKey'],
                            value=output['OutputValue'],
                        ))

            return stack_outputs

    def output(self, output_key, **kwargs):
        stack_outputs = self.outputs(**kwargs)

        for output in stack_outputs:
            if output['OutputKey'] == output_key:
                return output['OutputValue']

        print("Can't find output parameter {} in stack {} under {} profile".format(
            output_key,
            self.stack_name,
            self.profile_name
        ))

        return None

    def validate(self, details=False):
        s3 = self.client('s3')
        for template in ([self.template] + self.includes):

            if template in shared.store['validated_templates']:
                print('Template {} has been validated already'.format(
                    template
                ))
                continue
            else:
                shared.store['validated_templates'].append(template)

            temp_filename = "temp/%s-%s" % (uuid.uuid4(),
                                            os.path.basename(template))
            print("Uploading %s to temporary location s3://%s/%s" %
                  (template, self.s3_bucket, temp_filename))
            S3Transfer(s3).upload_file(
                template,
                self.s3_bucket,
                temp_filename,
                extra_args={'ACL': 'bucket-owner-full-control'}
            )

            template_url = "https://s3.amazonaws.com/%s/%s" % (
                self.s3_bucket, temp_filename)
            print("Validating template %s" % template_url)
            resp = self.client('cloudformation').validate_template(
                TemplateURL=template_url
            )

            if details:
                print('Template {} details: {}'.format(
                    template,
                    json.dumps(resp, indent=2, separators=(',', ': '))))

            print("Removing temporary file /%s from s3" % temp_filename)
            s3.delete_object(
                Bucket=self.s3_bucket,
                Key=temp_filename,
            )

    def create(self, **kwargs):
        self._upload()

        cloudformation = self.client('cloudformation')

        if 'stack_name' in kwargs:
            self.stack_name = kwargs.get('stack_name')

        template_url = "https://s3.amazonaws.com/%s/%s" % (
            self.s3_bucket, self.s3_key)
        print("Creating stack {}".format(self.stack_name))
        stack_id = None
        parameters = self._join_parameters(self.parameters, kwargs.get('parameters', None))
        try:
            resp = cloudformation.create_stack(
                StackName=self.stack_name,
                TemplateURL=template_url,
                Capabilities=['CAPABILITY_NAMED_IAM'],
                OnFailure=self.on_failure,
                Parameters=parameters
            )
            resp = cloudformation.describe_stacks(
                StackName=resp['StackId']
            )
            stack_id = resp['Stacks'][0]['StackId']
            timestamp = resp['Stacks'][0]['CreationTime']
        except botocore.exceptions.ParamValidationError as error:
            if str(error).startswith("Parameter validation failed:\nInvalid type for parameter"):
                res = re.search((r'Invalid type for parameter Parameters\[(\d)\].ParameterValue, '
                                r'value: None, type: (.+), valid types: (.+)'), str(error))
                param_idx = int(res.group(1))
                param_type_asis = res.group(2)
                param_type_tobe = res.group(3)

                print('Validation error:')
                print('Parameter {param_name} has invalid type/value.'.format(
                    param_name=parameters[param_idx]['ParameterKey']
                ))
                print('Expected: {expected}, Actual: {actual}/{value}'.format(
                    expected=param_type_tobe,
                    actual=param_type_asis,
                    value=parameters[param_idx]['ParameterValue']
                ))
                sys.exit(1)
            else:
                raise error
        except botocore.exceptions.ClientError as error:
            err_msg = error.response['Error']['Message']
            err_code = error.response['Error']['Code']
            if err_code == 'ValidationError':
                print('Validation error: {}'.format(err_msg))
                sys.exit(1)
            else:
                raise error

        self._print_events(stack_id, timestamp)

        self.outputs(no_cache=True, print=True)

        return

    def update(self, **kwargs):
        self._upload()

        cloudformation = self.client('cloudformation')

        if 'stack_name' in kwargs:
            self.stack_name = kwargs.get('stack_name')

        template_url = "https://s3.amazonaws.com/%s/%s" % (
            self.s3_bucket, self.s3_key)
        print("Updating stack {}".format(self.stack_name))
        timestamp = datetime.now(timezone.utc)
        resp = cloudformation.update_stack(
            StackName=self.stack_name,
            TemplateURL=template_url,
            Capabilities=['CAPABILITY_NAMED_IAM'],
            Parameters=self._join_parameters(
                self.parameters, kwargs.get('parameters', None))
        )

        self._print_events(resp['StackId'], timestamp)

        self.outputs(no_cache=True, print=True)

        return

    def delete(self, **kwargs):
        cloudformation = self.client('cloudformation')

        if 'stack_name' in kwargs:
            self.stack_name = kwargs.get('stack_name')

        resp = cloudformation.describe_stacks(StackName=self.stack_name)

        print('Deleting stack {}'.format(self.stack_name))
        timestamp = datetime.now(timezone.utc)
        cloudformation.delete_stack(
            StackName=resp['Stacks'][0]['StackId']
        )

        self._print_events(resp['Stacks'][0]['StackId'], timestamp)

        cache_key = 'cloudformation.outputs.{}.{}.{}'.format(
            self.region_name,
            self.profile_name,
            self.stack_name
        )
        self.cache(cache_key, list())

        return

    def wait_complete(self, **kwargs):
        cloudformation = self.client('cloudformation')

        stack_name = self.stack_name
        if 'stack_name' in kwargs:
            stack_name = kwargs.get('stack_name')

        print('Waiting stack {} status complete'.format(stack_name))
        waiter = cloudformation.get_waiter('stack_create_complete')
        waiter.wait(
            StackName=stack_name
        )

        return

    def wait_update(self, **kwargs):
        cloudformation = self.client('cloudformation')

        stack_name = self.stack_name
        if 'stack_name' in kwargs:
            stack_name = kwargs.get('stack_name')

        print('Waiting stack {} status updated'.format(stack_name))
        waiter = cloudformation.get_waiter('stack_update_complete')
        waiter.wait(
            StackName=stack_name
        )

        return

    def wait_delete(self, **kwargs):
        cloudformation = self.client('cloudformation')

        stack_name = self.stack_name
        if 'stack_name' in kwargs:
            stack_name = kwargs.get('stack_name')

        print('Waiting stack {} status deleted'.format(stack_name))
        waiter = cloudformation.get_waiter('stack_delete_complete')
        waiter.wait(
            StackName=stack_name
        )

        return

    def dry_run(self, **kwargs):
        self._upload()

        cloudformation = self.client('cloudformation')

        stack_name = self.stack_name
        if 'stack_name' in kwargs:
            stack_name = kwargs.get('stack_name')

        template_url = "https://s3.amazonaws.com/%s/%s" % (
            self.s3_bucket, self.s3_key)

        change_set_name = 'cs{ts}'.format(ts=time.strftime('%Y%m%d%H%M%S'))
        client_token = 'token{uuid}'.format(uuid=uuid.uuid4())

        print("Running dry_run for stack {}".format(stack_name))
        change_set = cloudformation.create_change_set(
            StackName=stack_name,
            ChangeSetName=change_set_name,
            ClientToken=client_token,
            ChangeSetType='UPDATE',
            TemplateURL=template_url,
            Capabilities=['CAPABILITY_NAMED_IAM'],
            Parameters=self._join_parameters(
                self.parameters,
                kwargs.get('parameters', None)
            )
        )
        waiter = cloudformation.get_waiter('change_set_create_complete')
        try:
            waiter.wait(
                StackName=stack_name,
                ChangeSetName=change_set_name
            )

            print('Changes to be performed:')
            next_token = None
            while True:
                args = {
                    'StackName': stack_name,
                    'ChangeSetName': change_set['Id'],
                }
                if next_token:
                    args['NextToken'] = next_token

                resp = cloudformation.describe_change_set(**args)

                for change in resp['Changes']:
                    print('{action}: {id} - {res_type}'.format(
                        action=change['ResourceChange']['Action'],
                        id=change['ResourceChange']['LogicalResourceId'],
                        res_type=change['ResourceChange']['ResourceType'],
                    ))

                if 'NextToken' not in resp:
                    break
                else:
                    next_token = resp['NextToken']

        except botocore.exceptions.WaiterError as error:
            if error.kwargs['reason'] == ('Waiter encountered a '
               'terminal failure state'):
                resp = cloudformation.describe_change_set(
                    StackName=stack_name,
                    ChangeSetName=change_set['Id'],
                )
                if resp['StatusReason'] == (
                    "The submitted information "
                    "didn't contain changes. "
                    "Submit different information "
                    "to create a change set."
                ):
                    print("No changes.")
                else:
                    raise Exception("Unknown error", None, sys.exc_info()[2])
            else:
                raise Exception("Unknown error", None, sys.exc_info()[2])
        finally:
            print('Deleting change set')
            cloudformation.delete_change_set(
                StackName=stack_name,
                ChangeSetName=change_set['Id']
            )

        return

    def estimate_cost(self, **kwargs):
        self._upload()

        cloudformation = self.client('cloudformation')

        stack_name = self.stack_name
        if 'stack_name' in kwargs:
            stack_name = kwargs.get('stack_name')

        template_url = "https://s3.amazonaws.com/{}/{}".format(
            self.s3_bucket, self.s3_key)
        print("Estimating template s3://{}/{}".format(
            self.s3_bucket, self.s3_key
        ))
        resp = cloudformation.estimate_template_cost(
            TemplateURL=template_url,
            Parameters=self._join_parameters(
                self.parameters, kwargs.get('parameters', None))
        )

        print('Check URL to see your template costs estimateion:\n{}'.format(
            resp['Url']))

        return

    def _upload(self):
        print("Uploading %s to s3://%s/%s" %
              (self.template, self.s3_bucket, self.s3_key))
        S3Transfer(self.client('s3')).upload_file(
            self.template,
            self.s3_bucket,
            self.s3_key,
            extra_args={'ACL': 'bucket-owner-full-control'}
        )

        s3_key = self.s3_key
        if not s3_key.endswith('/'):
            s3_key = s3_key[:s3_key.rfind('/') + 1]

        for file in (self.includes + self.resources):
            file_s3_key = '{}{}'.format(s3_key, os.path.basename(file))

            print("Uploading %s to s3://%s/%s" %
                  (file, self.s3_bucket, file_s3_key))
            S3Transfer(self.client('s3')).upload_file(
                file,
                self.s3_bucket,
                file_s3_key,
                extra_args={'ACL': 'bucket-owner-full-control'}
            )

    def _join_parameters(self, params1, params2):
        if (params1 and type(params1) != list) or \
           (params2 and type(params2) != list):
            raise Exception("Parameters argument should be a list() or None")

        if not params1 and params2:
            return params2
        elif params1 and not params2:
            return params1
        elif params1 and params2:
            result_d = dict()
            for param in params1:
                result_d[param['ParameterKey']] = param['ParameterValue']
            for param in params2:
                result_d[param['ParameterKey']] = param['ParameterValue']

            result = list()
            for key in result_d:
                result.append({
                    'ParameterKey': key,
                    'ParameterValue': result_d[key]
                })
            return result
        else:
            return list()

    def _trunc(self, text, length):
        if len(text) > length:
            ln_before = int(length / 2) - 2 + int(length / 4)
            ln_after = length - ln_before - 2
            result = text[:ln_before] + '..' + text[-ln_after:]
            return result
        else:
            return text.ljust(length)

    def _print_events(self, root_stack_id, start_ts):
        cloudformation = self.client('cloudformation')
        print("Printing events for {} and all it's child".format(root_stack_id))
        stack_ids = [root_stack_id]
        stack_data_template = {
            'event_ids_shown': set(),
            'stack_name': self.stack_name
        }
        stack_data = {
            root_stack_id: stack_data_template.copy()
        }

        paginator = cloudformation.get_paginator('describe_stack_events')

        root_stack_running = True
        stacks_running = 0
        last_out_msg = ''
        while(root_stack_running):
            for stack_id in stack_ids:
                try:
                    resp_iterator = paginator.paginate(StackName=stack_id)
                    for page in resp_iterator:
                        for event in page['StackEvents']:
                            phResId = event.get('PhysicalResourceId', None)
                            if event['EventId'] not in stack_data[stack_id]['event_ids_shown'] and \
                               event['Timestamp'] > start_ts:
                                if 'AWS::CloudFormation::Stack' == event['ResourceType'] and \
                                   phResId and \
                                   phResId not in stack_ids:
                                    stack_ids.insert(0, phResId)
                                    stack_data[phResId] = stack_data_template.copy()
                                    stack_data[phResId]['stack_name'] = event['LogicalResourceId']

                                stack_data[stack_id].get('event_ids_shown').add(event['EventId'])

                                logical_resource_id = event['LogicalResourceId']

                                if 'AWS::CloudFormation::Stack' == event['ResourceType'] and \
                                   event['StackId'] == phResId:
                                    logical_resource_id = stack_data[stack_id]['stack_name']

                                out_msg = ' / '.join((
                                    event['Timestamp'].strftime('%H:%m:%S'),
                                    self._trunc(stack_data[stack_id]['stack_name'], 16),
                                    self._trunc(logical_resource_id, 24),
                                    event['ResourceType'].replace('AWS::', '').ljust(37),
                                    event['ResourceStatus'].ljust(28),
                                    event.get('ResourceStatusReason', ''),
                                ))

                                if event['ResourceType'] == 'AWS::CloudFormation::Stack' and \
                                   event['ResourceStatus'] in STATUS_CF_RUN_COMPLETE and \
                                   event['PhysicalResourceId'] == root_stack_id:
                                    print('Stopping events printing on status {}. ({}, {})'.format(
                                        event['ResourceStatus'], stacks_running, len(stack_ids)
                                    ))
                                    root_stack_running = False
                                    stacks_running = stacks_running + 1
                                    last_out_msg = out_msg
                                else:
                                    print(out_msg)

                except botocore.exceptions.ClientError as err:
                    err_msg = err.response['Error']['Message']
                    err_code = err.response['Error']['Code']
                    if err_msg == 'Rate exceeded' and err_code == 'Throttling':
                        print('Rate exceeded. Forcing timeout.')
                        time.sleep(1)
                    else:
                        print('Msg: {}, Code: {}'.format(err_msg, err_code))
                        raise Exception(err)
        print(last_out_msg)



from setuptools import setup
import vulcan.meta_aws
setup(
    name='vulcan-aws',
    version=vulcan.meta_aws.__version__,
    author='Peter Salnikov',
    author_email='opensource@exrny.com',
    url=vulcan.meta_aws.__website__,
    packages=['vulcan', 'vulcan.aws', 'vulcan.aws.services'],
    install_requires=['boto3'],
    license='MIT License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: System Administrators',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3'
    ],
    keywords=['framework'],
    description='Amazon AWS boto3 helper libs.',
    long_description=open('README.rst').read()+'\n'+open('CHANGES.rst').read()
)
